#  Copyright 2022 New Vector Ltd
#
# SPDX-License-Identifier: AGPL-3.0-only OR LicenseRef-Element-Commercial
# Please see LICENSE files in the repository root for full details.

from matrix_content_scanner.utils.rust import check_rust_lib_up_to_date

check_rust_lib_up_to_date()
